import { h } from 'preact';
import { useState, useCallback, useEffect, useMemo } from 'preact/hooks';
import TextareaAutosize from 'react-textarea-autosize';
import PoissonDiskSampling from 'poisson-disk-sampling';
import './app.css';

const useHash = () => {
  const [hash, setHash] = useState(() => window.location.hash);

  const hashChangeHandler = useCallback(() => {
    setHash(window.location.hash);
  }, []);

  useEffect(() => {
    window.addEventListener('hashchange', hashChangeHandler);
    return () => {
      window.removeEventListener('hashchange', hashChangeHandler);
    };
  }, []);

  const updateHash = useCallback(
    (newHash: string) => {
      if (newHash !== hash) window.history.replaceState({}, '', `#${newHash}`);
    },
    [hash]
  );

  return [hash, updateHash] as const;
};

const SYMBOLS = [
  {
    symbol: "■",
    frequency: 3,
  }, {
    symbol: "▲",
    frequency: 12,
  }, {
    symbol: "●",
    frequency: 4,
  }, {
    symbol: "🟊",
    frequency: 2,
  }, {
    symbol: "🞧",
    frequency: 3,
  }, {
    symbol: "🗲",
    frequency: 2,
  },
];

function weightedRandomSymbol(): string {
  const total = SYMBOLS.map((s) => s.frequency).reduce((a, b) => a + b, 0);
  const sampleValue = Math.random() * total;

  let rollingValue = 0;
  for (const s of SYMBOLS) {
    rollingValue += s.frequency;
    if (rollingValue >= sampleValue) {
      return s.symbol;
    }
  }
  throw new Error("bad sample");
}

const SYMBOL_COUNT = 30;
const COLOR_COUNT = 20;

type SymbolData = {
  symbol: string;
  color: string;
  positionPx: [number, number];
  sizeEm: number;
  zIndex: number;
  innerAnimationDelaySec: number;
};

function SymbolBurst() {
  const getPoissonDiskSampler = () => new PoissonDiskSampling({
    shape: [window.innerHeight, window.innerWidth],
    minDistance: (window.innerWidth + window.innerHeight) / 20,
    maxDistance: (window.innerWidth + window.innerHeight) / 5,
    tries: 10,
  });

  const pds = getPoissonDiskSampler();

  const handleResizeEvent = (event: any) => getPoissonDiskSampler();

  useEffect(() => {
    addEventListener('resize', handleResizeEvent);
    return () => removeEventListener('resize', handleResizeEvent);
  }, []);

  const symbolData = useMemo(() => {
    const symbolData: SymbolData[] = [];
    const points = pds.fill();
    for (const point of points) {
      symbolData.push({
        symbol: weightedRandomSymbol(),
        color: `var(--ig-color-${1 + Math.floor(Math.random() * COLOR_COUNT)})`,
        positionPx: [point[0], point[1]],
        sizeEm: 1 + Math.random() * 2,
        zIndex: 2 * Math.floor(2 * Math.random()),
        innerAnimationDelaySec: -Math.random() * 30,
      });
    }
    return symbolData;
  }, [window.innerWidth, window.innerHeight]);

  return (
    <div class="symbol-burst">
      {symbolData.map((s) => (
        <div class="symbol" style={{
          color: s.color,
          top: `${s.positionPx[0]}px`,
          left: `${s.positionPx[1]}px`,
          fontSize: `${s.sizeEm}em`,
          zIndex: s.zIndex,
        }}>
          <span class={s.symbol} style={{ animationDelay: `${s.innerAnimationDelaySec}s` }}>{s.symbol}</span>
        </div>
      ))}
    </div>
  )
}

export function App() {
  const [hash, setHash] = useHash();
  const [title, setTitle] = useState(hash ? atob(decodeURIComponent(hash.substring(1))) : "i've gay");
  const [titleHeight, setTitleHeight] = useState(window.innerHeight * 0.8);
  const [titleRowHeight, setTitleRowHeight] = useState(titleHeight / 2);
  const [titleHeightCalculations, setTitleHeightCalculations] = useState(0);
  const [fontSizePx, setFontSizePx] = useState(48);

  const calculateFontSize = useCallback(() => {
    setFontSizePx((Math.min(window.innerWidth, window.innerHeight) * 0.5) / (titleHeight / titleRowHeight));
  }, [titleHeight, titleRowHeight, window.innerHeight]);

  const handleResizeEvent = (event: any) => {
    handleHeightChange(titleHeight, { rowHeight: titleRowHeight });
    calculateFontSize();
  }

  useEffect(() => {
    addEventListener('resize', handleResizeEvent);
    return () => removeEventListener('resize', handleResizeEvent);
  }, []);

  // Update hash when the title changes
  useEffect(() => {
    setHash(encodeURIComponent(btoa(title.trimEnd())));
  }, [title]);

  const handleTitleChange = (event: any) => {
    setTitle((event.target as HTMLTextAreaElement)?.value);
    setTitleHeightCalculations(0);
  }
  const handleHeightChange = useCallback((height: number, { rowHeight }: { rowHeight: number }) => {
    setTitleHeight(height);
    setTitleRowHeight(rowHeight);

    // Good height
    if (Math.abs(height / window.innerHeight - 0.8) < 0.05) return;
    // Acceptable height
    if (titleHeightCalculations >= 5 && height / window.innerHeight < 0.85) return;
    // Don't loop forever
    if (titleHeightCalculations >= 10) return;

    setTitleHeightCalculations((n) => n + 1);
    calculateFontSize();
  }, [titleHeightCalculations]);

  console.log(`fontSize: ${fontSizePx}px; titleHeight: ${titleHeight}; titleRowHeight: ${titleRowHeight}; titleHeightCalculations: ${titleHeightCalculations}`);

  return (
    <div class="ivegay">
      <TextareaAutosize
        value={title}
        minRows={2}
        style={{ fontSize: `${fontSizePx}px` }}
        onChange={handleTitleChange}
        onHeightChange={handleHeightChange}
        spellCheck={false}
        ref={(tag) => (tag ? tag.spellcheck = false : null)}
      />
      <SymbolBurst />
      <p>
        Click to edit · <a href="https://www.teepublic.com/t-shirt/1810231-ive-gay">Shirt</a> · <a href="https://diverse.direct/c-h-s/chsep-0003/">E.P.</a> · <a href="https://meow.garden/">Made by Ash</a>
      </p>
    </div>
  )
}
